;  ------------------------------------------------------------------

;  Program to manage data entry into a four-page multi-table form.
;  The program and form make use of a special master table with
;  linked fields to achieve various effects.

;  ------------------------------------------------------------------

Clear
@ 10,0 ?? "One moment please..."

;  ------------------------------------------------------------------

Proc Main()

;  The main procedure for the data entry program.

;  Initialization:  View the master form, and create an empty
;  record in it.

Clear                   ;clear the screen
@ 10,0 ?? "Initializing..."
Reset                   ;clear the workspace

Lock "Master" PWL,      ;place the necessary table locks for multi-user
     "Student" PWL,
     "Class" PWL,
     "ShadStu" PWL,
     "ShadClas" PWL,
     "Enroll1" PWL,
     "Enroll2" PWL,
     "Scores" PWL
If not retval then
   Quit "Could not lock tables:  "+errormessage()
Endif

CoEdit "Master"         ;view the master table, enter co-edit mode
Ins                     ;create a new blank record to work with
PickForm "F"            ;bring up the form
MoveTo "ShadStu"        ;insert a shadow record into the two shadow tables
Ins
MoveTo "ShadClas"
Ins

;  Allow the user to move from table to table within the form.

Table_Loop()            ;proc with the inter-table loop

;  All done

Clear                   ;clear the screen
MoveTo "ShadStu"        ;delete the shadow record from the two shadow tables
Del
MoveTo "ShadClas"
Del
MoveTo "Master"         ;return to the master table
Del                     ;delete the record we were working on
Do_it!                  ;end editing of the table
Reset                   ;clear the workspace

EndProc

;  ------------------------------------------------------------------

Proc Table_Loop()

;  This procedure manages the movement from one table to another
;  on the form, as the user moves through them.

Private
   be_proc,          ;begin-edit procedure
   class_rec,        ;the record for the current class
   current_student,  ;the current student ID
   current_class,    ;the current class ID
   current_page,     ;the current page # on the form
   current_table,    ;the name of the current table
   de_proc,          ;record-deletion procedure
   del_key,          ;function key to delete a record
   down_key,         ;function key to move down a record
   end_key,          ;function key to move to the last record
   home_key,         ;function key to move to the first record
   image_mode,       ;which ImageRights mode to use:  ReadOnly, Update, All
   ins_key,          ;function key to insert a new record
   is_editing,       ;whether the user is in "edit" mode on the table
   next_page,        ;the next page # on the form
   next_table,       ;the name of the next table to move to
   pgup_key,         ;function key to move up a page
   pgdn_key,         ;function key to move down a page
   prompt1,prompt2,  ;two lines of messages
   prompt2_edit,     ;prompt to display while editing
   prompt2_not_edit, ;prompt to display while not editing
   ra_proc,          ;record-arrival procedure
   rcd_mvmt_type,    ;type of record movement performed on a table
   rd_proc,          ;record-departure procedure
   rm_proc,          ;record-movement procedure
   student_rec,      ;the record for the current student
   up_key,           ;function key to move up a record
   user_key,         ;the key pressed by user during Wait Record
   wait_msg,         ;a message to display to user during Wait
   zoom_key,         ;function key to locate a record (zoom)
   znxt_key          ;function key to locate the next record (zoom-next)

;  Initialization

current_page = 1           ;no page has been initialized yet
next_page = 1              ;we'll start on page 1
next_table = "STUDENT"     ;this is the first table to move to
current_student = ""       ;no student has been processed yet
Array student_rec[1]       ;no student record array yet
student_rec[1]  = ""       ;no student record in an array yet
current_class = ""         ;no class has been processed yet
Array class_rec[1]         ;no class record array yet
class_rec[1]  = ""         ;no class record in an array yet
wait_msg = ""              ;no message to display yet

;  Move from table to table in response to the user's key-presses.

While True                          ;one loop for each table movement
   If next_page<>current_page then  ;we're moving to a new page
      Page_Arrival()                ;perform arrival processing for the page
   Endif
   MoveTo_Table()                   ;move to the next table
   Table_Arrival()                  ;init params based on current table
   Record_Loop(ra_proc,rd_proc,     ;proc with the inter-record loop
               be_proc,rm_proc,de_proc)
   Which_Table_Next()               ;determine next table to move to
   If next_table="" then            ;the user has pressed [F2] to exit
      QuitLoop                      ;exit from this program
   Endif
   Table_Departure()                ;processing before leaving this table
EndWhile 

EndProc

;  ------------------------------------------------------------------

Proc Page_Arrival()

;  Perform page-arrival processing.

Switch
   Case next_page=2:    Page_Arrival_2()
   Case next_page=4:    Page_Arrival_4()
EndSwitch

current_page = next_page

EndProc

;  ------------------------------------------------------------------

Proc Page_Arrival_2()

;  Perform page-arrival processing for page 2.

If isblank(current_student) then    ;we don't have a student yet
   MoveTo "Student"                 ;go to the student table
   current_student = [Student ID]   ;find out the current student
   CopyToArray student_rec          ;fetch the entire student record
Endif
If isblank(student_rec[1]) then     ;the student array is out-of-date
   MoveTo "Student"                 ;go to the student table
   MoveTo [Student ID]              ;the field to locate on
   If []<>current_student then      ;we're not on the right record
      Locate current_student        ;find the student record
      current_student = []          ;find out the current student
   Endif
   CopyToArray student_rec          ;fetch the entire student record
Endif

MoveTo "Master"                     ;we need to update the master record...
[Student ID] = current_student      ;...with the student # key
MoveTo "Shadstu"                    ;we need to update the shadow record
Del                                 ;delete the previous shadow record
Ins                                 ;open up a new shadow record
CopyFromArray student_rec           ;get the record from the source table

EndProc

;  ------------------------------------------------------------------

Proc Page_Arrival_4()

;  Perform page-arrival processing for page 4.

If isblank(current_class) then      ;we don't have a class yet
   MoveTo "Class"                   ;go to the class table
   current_class = [Class ID]       ;find out the current class
   CopyToArray class_rec            ;fetch the entire class record
Endif
If isblank(class_rec[1]) then       ;the class array is out-of-date
   MoveTo "Class"                   ;go to the class table
   MoveTo [Class ID]                ;the field to locate on
   If []<>current_class then        ;we're not on the right record
      Locate current_class          ;find the class record
      current_class = []            ;find out the current class
   Endif
   CopyToArray class_rec            ;fetch the entire class record
Endif
MoveTo "Master"                     ;we need to update the master record...
[Class ID] = current_class          ;...with the class # key
MoveTo "Shadclas"                   ;we need to update the shadow record
Del                                 ;delete the previous shadow record
Ins                                 ;open up a new shadow record
CopyFromArray class_rec             ;get the record from the source table

EndProc

;  ------------------------------------------------------------------

Proc MoveTo_Table()

;  Move to the next table on the form to be processed.

MoveTo next_table                   ;move to the specified table
current_table = upper(table())      ;remember its name
If current_table<>next_table then   ;we weren't able to move to it
   Reset
   Quit "Could not move to table "+next_table
Endif

EndProc

;  ------------------------------------------------------------------

Proc Table_Arrival()

;  Perform table-arrival processing.

image_mode = "ReadOnly"          ;start out with read-only mode
is_editing = False               ;start out not editing
prompt1 = "Shift-F1=Page 1, Shift-F2=Page 2, " +
          "Shift-F3=Page 3, Shift-F4=Page 4"
prompt2_edit     = "F2=End Editing,  F2+F2=Exit"
prompt2_not_edit = "F2=Exit,  F9=Edit"
prompt2 = prompt2_not_edit

rcd_mvmt_type = "Regular"
ra_proc = "Rec_Arr_Std"
rd_proc = "Rec_Dep_Std"
be_proc = ""
de_proc = ""
rm_proc = ""

Switch
   Case current_table="STUDENT":
   Case current_table="CLASS":
   Case current_table="SHADSTU":
      rcd_mvmt_type = "Special"
      ra_proc = "Rec_Arr_Shad_Stu"
      rd_proc = "Rec_Dep_Shad_Stu"
      rm_proc = "Move_Shad_Stu"
   Case current_table="SHADCLAS":
      rcd_mvmt_type = "Special"
      ra_proc = "Rec_Arr_Shad_Clas"
      rd_proc = "Rec_Dep_Shad_Clas"
      rm_proc = "Move_Shad_Clas"
   Case current_table="ENROLL1":
      ra_proc = "Rec_Arr_Enroll1"
      be_proc = "Bgn_Ed_Enroll1"
      rd_proc = "Rec_Dep_Enroll1"
      de_proc = "Rec_Del_Enroll1"
   Case current_table="ENROLL2":
      ra_proc = "Rec_Arr_Enroll2"
      be_proc = "Bgn_Ed_Enroll2"
      rd_proc = "Rec_Dep_Enroll2"
      de_proc = "Rec_Del_Enroll2"
   Case current_table="SCORES":
   Otherwise:
      Reset
      Quit "This table "+current_table+" is not handled by the program"
EndSwitch

Switch
   Case rcd_mvmt_type = "Regular":
      up_key   = "Up"
      down_key = "Down"
      home_key = "Home"
      end_key  = "End"
      pgup_key = "PgUp"
      pgdn_key = "PgDn"
      ins_key  = "Ins"
      del_key  = "Del"
      zoom_key = "Zoom"
      znxt_key = "ZoomNext"
   Case rcd_mvmt_type = "Special":
      up_key   = 0
      down_key = 0
      home_key = "Home"
      end_key  = "End"
      pgup_key = "PgUp"
      pgdn_key = "PgDn"
      ins_key  = "Ins"
      del_key  = "Del"
      zoom_key = 0
      znxt_key = 0
EndSwitch

EndProc

;  ------------------------------------------------------------------

Proc Which_Table_Next()

;  Determine which table to move to next, based on the function
;  key pressed by the user

Switch
   Case user_key="F2":           ;the user has pressed [F2]
      next_page = current_page   ;no page movement
      next_table = ""            ;there is no next table, we're done
   Case user_key="F11":          ;Shift-F1:  Move to page 1
      next_page = 1              ;the next page will be page 1
      next_table = "STUDENT"     ;this is the main table on page 1
   Case user_key="F12":          ;Shift-F2:  Move to page 2
      next_page = 2              ;the next page will be page 2
      next_table = "SHADSTU"     ;this is the main table on page 2
   Case user_key="F13":          ;Shift-F3:  Move to page 3
      next_page = 3              ;the next page will be page 3
      next_table = "CLASS"       ;this is the main table on page 3
   Case user_key="F14":          ;Shift-F4:  Move to page 4
      next_page = 4              ;the next page will be page 4
      next_table = "SHADCLAS"    ;this is the main table on page 4
   Case user_key="CtrlPgDn":           ;Ctrl-PgDn:  Move to next page
      If current_page<4 then           ;we're not on the last page yet
         next_page = current_page+1    ;move to the next page
         next_table = "Top of Page"    ;figure out the table from the page
      Else                             ;we're already on the last page
         next_page = current_page      ;so stay here
         next_table = current_table
      Endif
   Case user_key="CtrlPgUp":           ;Ctrl-PgUp:  Move to previous page
      If current_page>1 then           ;we're not on the first page yet
         next_page = current_page-1    ;move to the previous page
         next_table = "Top of Page"    ;figure out the table from the page
      Else                             ;we're already on the last page
         next_page = current_page      ;so stay here
         next_table = current_table
      Endif
   Case user_key="F3" or user_key="F4":      ;movement within a page
      next_page = current_page         ;no page movement
      KeyPress user_key                ;make user's table movement
      next_table = upper(table())      ;what table is it?
      If next_table="MASTER" then      ;don't allow mvmt to master
         KeyPress user_key             ;move past it
         next_table = upper(table())   ;the next table
      Endif
      MoveTo current_table             ;return to original table
EndSwitch

If next_table="Top of Page" then    ;we want to move to the top of new page
   Switch
      Case next_page=1:    next_table = "STUDENT"
      Case next_page=2:    next_table = "SHADSTU"     
      Case next_page=3:    next_table = "CLASS"       
      Case next_page=4:    next_table = "SHADCLAS"    
   EndSwitch
Endif

EndProc

;  ------------------------------------------------------------------

Proc Table_Departure()

;  Perform table-departure processing.

Switch
   Case current_table="STUDENT":
      current_student = [Student ID]
      CopyToArray student_rec
   Case current_table="ENROLL1":
      If not isblank([Class ID]) then
         current_class = [Class ID]       ;remember the latest class
         MoveTo "Class"                   ;go to the class table
         MoveTo [Class ID]                ;we'll search on the key field
         If []<>current_class then        ;we're not on the same class
            Locate current_class          ;try to find it
         Endif
         current_class = []               ;find out the current class
         CopyToArray class_rec            ;fetch the entire class record
      Endif
   Case current_table="CLASS":
      current_class = [Class ID]
      CopyToArray class_rec
   Case current_table="ENROLL2":
      If not isblank([Student ID]) then
         current_student = [Student ID]   ;remember the latest student
         MoveTo "Student"                 ;go to the student table
         MoveTo [Student ID]              ;we'll search on the key field
         If []<>current_student then      ;we're not on the same student
            Locate current_student        ;try to find it
         Endif
         current_student = []             ;find out the current student
         CopyToArray student_rec          ;fetch the entire student record
      Endif
EndSwitch

EndProc

;  ------------------------------------------------------------------

Proc Record_Loop(rec_arrive_proc, rec_depart_proc, 
                 begin_edit_proc, rec_move_proc, rec_del_proc)

;  This procedure manages the movement from one record to another
;  on the form, as the user moves around in a record

;  Note:  In my articles in Data Based Advisor, Jan/March/April 1991,
;         I presented many important tips for writing WAIT RECORD
;         procedures.  For brevity, I have omitted many of those tips
;         from this procedure.  However, you should refer to them
;         and follow them in your own programming.

;  Parameters:
;     rec_arrive_proc   -- procedure to perform upon record arrival
;     rec_depart_proc   -- procedure to perform upon record departure
;     begin_edit_proc   -- procedure to perform when beginning edit mode
;     rec_move_proc     -- procedure to perform movement
;     rec_del_proc      -- procedure to perform record deletion

Private
   is_arriving,   ;whether we're arriving at a new record
   is_posted_key  ;whether a new key has been posted yet

;  Initialization

is_arriving = True         ;it is a new record arrival the first time

;  Move from record to record in response to the user's key-presses.

While True                 ;one loop for each record movement

   If is_arriving then        ;we're arriving at a new record
      is_arriving = False     ;not a new record any more
      If not isblank(rec_arrive_proc) then
         ExecProc rec_arrive_proc         ;record arrival procedure
      Endif
   Endif

   Switch
      Case image_mode="ReadOnly":      ImageRights ReadOnly
      Case image_mode="Update":        ImageRights Update
   EndSwitch

   Wait Record
      Prompt prompt1, prompt2
      Message wait_msg
      Until "F2",                                  ;termination key
            "F9",                                  ;mode-change key
            "F3","F4","F11","F12","F13","F14",     ;table-mvmt keys
            "CtrlPgUp","CtrlPgDn",                 ;table-mvmt keys
            "Dos","DosBig","LockKey","KeyLookup",  ;illegal keys
            up_key,down_key,pgup_key,pgdn_key,     ;record-mvmt keys
            home_key,end_key,ins_key,del_key,
            zoom_key,znxt_key

   user_key = retval       ;remember which key the user pressed
   wait_msg = ""           ;clear the previous error message
   ImageRights             ;restore full image rights

   If user_key="Dos" or user_key="DosBig" or
      user_key="LockKey" or user_key="KeyLookup" then
         Beep Loop         ;illegal key
   Endif

   If user_key="Del" then
      Del_Confirm()                    ;ask user to confirm the [Del]
      If not retval then Loop Endif    ;don't delete the record
   Endif

   If is_editing and not isblank(rec_depart_proc) then
      ExecProc rec_depart_proc            ;record departure procedure
      If not retval then Loop Endif       ;don't proceed
   Endif

   If (user_key="F2" and not is_editing) or
      user_key="F3" or user_key="F4" or user_key="F11" or
      user_key="F12" or user_key="F13" or user_key="F14" or
      user_key="CtrlPgUp" or user_key="CtrlPgDn" then
         If rcd_mvmt_type="Regular" then
            If recordstatus("Locked") or
               (recordstatus("New") and recordstatus("Modified")) then
                  UnLockRecord
            Endif
         Endif
         QuitLoop          ;table-movement key
   Endif

   If user_key="F2" or user_key="F9" then          ;mode change key
      Switch
         Case user_key="F9" and is_editing:
            Beep
         Case user_key="F9":
            If not isblank(begin_edit_proc) then   ;editing handled spec'ly
               ExecProc begin_edit_proc            ;begin-edit procedure
            Else
               Begin_Edit()                        ;begin edit mode
            Endif
         Case user_key="F2":
            End_Edit()                             ;end edit mode
      EndSwitch
      Loop
   Endif

   Switch
      Case user_key="Del" and not isblank(rec_del_proc):
         ExecProc rec_del_proc            ;special deletion procedure
      Case not isblank(rec_move_proc):    ;movement is handled specially
         ExecProc rec_move_proc           ;record movement procedure
      Otherwise:
         If not is_editing then           ;we're in a display-only mode
            ImageRights ReadOnly          ;that affects how some keys work
         Endif
         Switch
            Case user_key="Zoom":
               Zoom_Echo(True)
            Otherwise:
               KeyPress user_key          ;simply press the key
         EndSwitch
         ImageRights                      ;restore full image rights
   EndSwitch
   is_arriving = True      ;we've arrived at a new record
EndWhile

EndProc

;  ------------------------------------------------------------------

Proc Begin_Edit()

;  Switch from read-only mode to edit (read/write) mode.
;  The program does not switch between Paradox's "Main" and "Edit" mode.
;  Instead, it remains in "CoEdit" mode the entire time,
;  and uses the ImageRights command to switch between read-only
;  and read/write.

is_editing = True                ;allow editing
is_arriving = True               ;act as if we've arrived at a new record
prompt2 = prompt2_edit

EndProc

;  ------------------------------------------------------------------

Proc Bgn_Ed_Enroll1()

;  Special procedure to begin editing the enrollment (#1) table.
;  This is like normal begin-editing, except we must also lock
;  the student record, to make sure no one else is also working
;  on the same student at this time.

MoveTo "Student"                          ;return to the student record
LockRecord                                ;lock the record
If not retval and errorcode()<>55 then    ;we couldn't lock it
   wait_msg = errormessage()              ;the error message
   MoveTo "Enroll1"                       ;return to the current table
   Return
Endif

MoveTo "Enroll1"                 ;return to the current table
is_editing = True                ;allow editing
is_arriving = True               ;act as if we've arrived at a new record
prompt2 = prompt2_edit

EndProc

;  ------------------------------------------------------------------

Proc Bgn_Ed_Enroll2()

;  Special procedure to begin editing the enrollment (#2) table.
;  This is like normal begin-editing, except we must also lock
;  the class record, to make sure no one else is also working
;  on the same class at this time.

MoveTo "Class"                            ;return to the class record
LockRecord                                ;lock the record
If not retval and errorcode()<>55 then    ;we couldn't lock it
   wait_msg = errormessage()              ;the error message
   MoveTo "Enroll2"                       ;return to the current table
   Return
Endif

MoveTo "Enroll2"                 ;return to the current table
is_editing = True                ;allow editing
is_arriving = True               ;act as if we've arrived at a new record
prompt2 = prompt2_edit

EndProc

;  ------------------------------------------------------------------

Proc End_Edit()

;  Switch from edit (read/write) mode to read-only mode.

is_editing = False               ;curtail editing
image_mode = "ReadOnly"          ;switch the table to read-only mode
prompt2 = prompt2_not_edit

EndProc

;  ------------------------------------------------------------------

Proc Rec_Arr_Std()

;  Standard record-arrival procedure during editing

If not is_editing then
   Return                  ;this stuff only applies while editing
Endif

image_mode = "ReadOnly"    ;assume the user can't change the record
is_posted_key = True       ;assume this is an existing key

If not recordstatus("New") and         ;it's an existing record
   not recordstatus("Locked") then     ;but not yet locked
      LockRecord                       ;attempt to lock the record
      If not retval then               ;the lock failed
         wait_msg = errormessage()     ;display the error message
      Endif
Endif

If recordstatus("New") or           ;it's a new record, or
   recordstatus("Locked") then      ;it's a locked record
      If recordstatus("New") then   ;it's a new record
         CtrlHome                   ;always start at the leftmost field
         image_mode = "All"         ;the user must be able to change all
         is_posted_key = False      ;we haven't saved the key yet
      Else                          ;it's an existing record
         image_mode = "Update"      ;let the user modify it
      Endif
Endif

EndProc

;  ------------------------------------------------------------------

Proc Rec_Arr_Shad_Stu()

;  Record-arrival procedure during editing shadow students

Rec_Arr_Page2()            ;update links for the class-name display

If not is_editing then
   Return                  ;this stuff only applies while editing
Endif

MoveTo "Student"           ;move to the source table, to lock the record
Rec_Arr_Std()              ;perform standard record arrival (try to lock)
MoveTo "ShadStu"           ;return to the shadow table
If not is_posted_key then  ;it's a new record
   CtrlHome                ;start at the left margin
Endif

EndProc

;  ------------------------------------------------------------------

Proc Rec_Arr_Shad_Clas()

;  Record-arrival procedure during editing shadow classes

Rec_Arr_Page4()            ;update links for the test-score display

If not is_editing then
   Return                  ;this stuff only applies while editing
Endif

MoveTo "Class"             ;move to the source table, to lock the record
Rec_Arr_Std()              ;perform standard record arrival (try to lock)
MoveTo "ShadClas"          ;return to the shadow table
If not is_posted_key then  ;it's a new record
   CtrlHome                ;start at the left margin
Endif

EndProc

;  ------------------------------------------------------------------

Proc Rec_Arr_Enroll1()

;  Record-arrival procedure during editing enrollments by student

;  It is necessary to locate and lock the corresponding 
;  class record.

Rec_Arr_Page2()               ;update links for the class-name display

If not is_editing then
   Return                     ;this stuff only applies while editing
Endif

image_mode = "ReadOnly"       ;assume the user can't change the record
is_posted_key = True          ;assume this is an existing key

;  Change the current class in the master table

If not recordstatus("New") then     ;this is an existing record
   current_class = [Class ID]       ;the class for the current record
   class_rec[1]  = ""               ;no class record in an array yet
   If isblank(current_class) then   ;there isn't a class
      wait_msg = "Can't edit this record; no Class ID"
      Return
   Endif
   MoveTo "Master"                  ;in the dummy master record
   [Class ID] = current_class       ;fill in the class

;  Lock the class record

   MoveTo "Class"
   If [Class ID]<>current_class then      ;we're not on it yet
      MoveTo [Class ID]
      Locate current_class
      If not retval then
         wait_msg = "No such class as "+current_class
         MoveTo "Enroll1"
         Return
      Endif
   Endif
   If not recordstatus("Locked") then     ;it's not locked
      LockRecord                          ;attempt to lock the record
      If not retval then                  ;the lock failed
         wait_msg = errormessage()        ;display the error message
      Endif
   Endif
   MoveTo "Enroll1"           ;move to the table, to lock the record
Endif

;  Lock the enrollment record

Rec_Arr_Std()              ;perform standard record arrival (try to lock)

EndProc

;  ------------------------------------------------------------------

Proc Rec_Arr_Enroll2()

;  Record-arrival procedure during editing enrollments by class

;  It is necessary to locate and lock the corresponding 
;  student record.

Rec_Arr_Page4()               ;update links for the test-score display

If not is_editing then
   Return                     ;this stuff only applies while editing
Endif

image_mode = "ReadOnly"       ;assume the user can't change the record
is_posted_key = True          ;assume this is an existing key

;  Change the current student in the master table

If not recordstatus("New") then     ;this is an existing record
   current_student = [Student ID]   ;the student for the current record
   student_rec[1] = ""              ;the record array is out of date
   If isblank(current_student) then ;there isn't a student
      wait_msg = "Can't edit this record; no Student ID"
      Return
   Endif
   MoveTo "Master"                  ;in the dummy master record
   [Student ID] = current_student   ;fill in the student

;  Lock the student record

   MoveTo "Student"
   If [Student ID]<>current_student then  ;we're not on it yet
      MoveTo [Student ID]
      Locate current_student
      If not retval then
         wait_msg = "No such student as "+current_student
         MoveTo "Enroll2"
         Return
      Endif
   Endif
   If not recordstatus("Locked") then     ;it's not locked
      LockRecord                          ;attempt to lock the record
      If not retval then                  ;the lock failed
         wait_msg = errormessage()        ;display the error message
      Endif
   Endif
   MoveTo "Enroll2"           ;move to the table, to lock the record
Endif

;  Lock the enrollment record

Rec_Arr_Std()              ;perform standard record arrival (try to lock)

EndProc

;  ------------------------------------------------------------------

Proc Rec_Arr_Page2()

;  Update the linking fields necessary to update the display tables
;  on page 2 of the form.

Private
   detail_recno,     ;record # of the detail record
   is_new_detail     ;whether it's a new detail record

;  Move to the ENROLL1 table, which has the [Class ID] field which
;  we need to capture

is_new_detail = False         ;assume it's not a new detail record
If current_table<>"ENROLL1" then
   MoveTo "Enroll1"
Else
   If recordstatus("New") and not recordstatus("Modified") then
      is_new_detail = True    ;a new detail record will vanish when we move
      detail_recno = recno()  ;remember the record number
   Endif
Endif

;  Transfer the [Class ID] to the master table.  At that point,
;  Paradox will fill the [Class Name] in automatically, owing to the
;  TableLookup-Fill-In feature.

current_class = [Class ID]
class_rec[1]  = ""         ;no class record in an array yet
MoveTo "Master"
[Class ID] = current_class
If isblank(current_class) then
   [Class Name] = ""
Endif
MoveTo current_table

;  If a newly inserted record vanished as a result of this table movement,
;  replace it.

If is_new_detail then                  ;there was a newly inserted record
   If not recordstatus("New") then     ;but it has disappeared
      If recno()=detail_recno then     ;we're still at the same spot
         Ins                           ;so re-insert the blank record
      Else                             ;we moved
         End Down                      ;so re-add blank record at the bottom
      Endif
   Endif
Endif

EndProc

;  ------------------------------------------------------------------

Proc Rec_Arr_Page4()

;  Update the linking fields necessary to update the display tables
;  on page 4 of the form.

Private
   detail_recno,     ;record # of the detail record
   is_new_detail     ;whether it's a new detail record

;  Move to the ENROLL2 table, which has the [Student ID] field which
;  we need to capture

is_new_detail = False         ;assume it's not a new detail record
If current_table<>"ENROLL2" then
   MoveTo "Enroll2"
Else
   If recordstatus("New") and not recordstatus("Modified") then
      is_new_detail = True    ;a new detail record will vanish when we move
      detail_recno = recno()  ;remember the record number
   Endif
Endif

;  Transfer the [Student ID] to the master table.  At that point,
;  Paradox will fill the student name in automatically, owing to the
;  TableLookup-Fill-In feature.  Paradox will also automatically update
;  the related records displayed in the SCORES table, owing to the
;  linked multi-table forms feature.

current_student = [Student ID]
student_rec[1] = ""              ;the record array is out of date
MoveTo "Master"
[Student ID] = current_student
If isblank(current_student) then
   [First Name] = ""
   [Last Name]  = ""
Endif
MoveTo current_table

;  If a newly inserted record vanished as a result of this table movement,
;  replace it.

If is_new_detail then                  ;there was a newly inserted record
   If not recordstatus("New") then     ;but it has disappeared
      If recno()=detail_recno then     ;we're still at the same spot
         Ins                           ;so re-insert the blank record
      Else                             ;we moved
         End Down                      ;so re-add blank record at the bottom
      Endif
   Endif
Endif

EndProc

;  ------------------------------------------------------------------

Proc Rec_Dep_Std()

;  Standard record-departure procedure during editing

Private
   orig_field        ;the field on which the user has placed the cursor

;  If the user presses the [Del] key, there is no special processing.
;  We simply delete the record.

If user_key="Del" then     ;the [Del] key
   Return True             ;no special processing or testing to do
Endif

;  Make sure the user has fixed a prior duplicate key

If recordstatus("KeyViol") then     ;we still have dup key
   Beep                             ;display an error message
   wait_msg = "Fix the duplicate key!"
   CtrlHome    ;move the cursor to where key field usually is
   Return False                     ;do not proceed
Endif

;  Check whether a new record has a duplicate key.
;  Note:  This procedure takes advantage of the fact that for every
;  table in the sample form, the user fills in exactly one key field
;  and that key field is the leftmost field in the form.  If this were
;  not the case, parts of this code would have to be modified.

If not is_posted_key and               ;new key not yet posted
   recordstatus("Modified") then       ;user changed the record
      orig_field = field()             ;remember the current field
      CtrlHome                         ;move to the key field
      If isblank([]) then              ;the key hasn't been filled in
         Beep                          ;indicate trouble
         wait_msg = "You must fill in this field"
         Return False                  ;do not proceed
      Endif
      LockRecord                       ;attempt to lock it
      If not retval then               ;the lock failed
         Beep                          ;indicate trouble
         If errorcode()=53 then        ;duplicate key error
            wait_msg = "Another record with this key already exists"
         Else
            wait_msg = errormessage()  ;some other locking error
         Endif
         Return False                  ;do not proceed
      Endif
      MoveTo Field orig_field          ;return to the original field
      is_posted_key = True             ;the key is valid
      image_mode = "Update"            ;allow no further key changes
Endif

Return True          ;successful completion

EndProc

;  ------------------------------------------------------------------

Proc Rec_Dep_Shad_Stu()

;  Record-departure procedure during editing shadow students

;  If the user presses the [Del] key, there is no special processing.
;  We simply delete the record.

If user_key="Del" then     ;the [Del] key
   Return True             ;no special processing or testing to do
Endif

;  Transfer the user's record back into the source table

CopyToArray student_rec    ;capture the record
MoveTo "Student"           ;the source table
CopyFromArray student_rec  ;copy the record into the source table

;  Perform standard record departure on the source record

Rec_Dep_Std()              ;standard record departure procedure
If not retval then         ;it failed
   MoveTo "ShadStu"        ;return to the shadow tabe
   CtrlHome                ;move to the key field
   Return False            ;do not proceed
Endif

;  Update the master table, and thus any other links,
;  with the new source table key

current_student = [Student ID]      ;find out the current student
MoveTo "Master"                     ;we need to update the master record...
[Student ID] = current_student      ;...with the student # key
MoveTo "Shadstu"                    ;we need to update the shadow record

Return True          ;successful completion

EndProc

;  ------------------------------------------------------------------

Proc Rec_Dep_Shad_Clas()

;  Record-departure procedure during editing shadow classes

;  If the user presses the [Del] key, there is no special processing.
;  We simply delete the record.

If user_key="Del" then     ;the [Del] key
   Return True             ;no special processing or testing to do
Endif

;  Transfer the user's record back into the source table

CopyToArray class_rec      ;capture the record
MoveTo "Class"             ;the source table
CopyFromArray class_rec    ;copy the record into the source table

;  Perform standard record departure on the source record

Rec_Dep_Std()              ;standard record departure procedure
If not retval then         ;it failed
   MoveTo "ShadClas"       ;return to the shadow tabe
   CtrlHome                ;move to the key field
   Return False            ;do not proceed
Endif

;  Update the master table, and thus any other links,
;  with the new source table key

current_class = [Class ID]          ;find out the current class
MoveTo "Master"                     ;we need to update the master record...
[Class ID] = current_class          ;...with the class # key
MoveTo "ShadClas"                   ;we need to update the shadow record

Return True          ;successful completion

EndProc

;  ------------------------------------------------------------------

Proc Rec_Dep_Enroll1()

;  Record-departure procedure during editing enrollments by student

Private
   enr_rec,       ;the record transferred between the two tables
   is_changed     ;whether the user has changed the record

;  Changes made to the ENROLL1 record must also be made to the
;  companion ENROLL2 record.

;  If the user presses the [Del] key, there is no special processing.
;  We simply delete the record.

If user_key="Del" then     ;the [Del] key
   Return True             ;no special processing or testing to do
Endif

;  Perform standard record departure on the record

is_changed = recordstatus("Modified")  ;has the user made changes?
Rec_Dep_Std()              ;standard record departure procedure
If not retval then         ;it failed
   Return False            ;do not proceed
Endif
If not is_posted_key or not is_changed then  ;is there some new change?
   Return True                               ;no, nothing more to do
Endif

;  Update the master table, and thus any other links,
;  with the new source table key

CopyToArray enr_rec              ;capture the new or changed record
current_class = [Class ID]       ;find out the current class
class_rec[1]  = ""               ;no class record in an array yet
MoveTo "Master"                  ;we need to update the master record...
[Class ID] = current_class       ;...with the class # key

;  Transfer the record to the ENROLL2 table

MoveTo "Enroll2"                             ;move to the companion table
enr_rec["Class ID"] = enr_rec["Student ID"]  ;swap the two key fields
enr_rec["Student ID"] = current_class
Ins                                          ;insert a new record
CopyFromArray enr_rec                        ;fill in the record
LockRecord                                   ;post it
If not retval then                           ;the lock failed
   If errorcode()<>53 then                   ;not a duplicate key
      wait_msg = "Error updating the ENROLL2 table:  "+errormessage()
      MoveTo "Enroll1"
      Return False
   Endif
Endif
UnLockRecord                                 ;post the changes

MoveTo "Enroll1"     ;return to the original table

Return True          ;successful completion

EndProc

;  ------------------------------------------------------------------

Proc Rec_Dep_Enroll2()

;  Record-departure procedure during editing enrollments by student

Private
   enr_rec,       ;the record transferred between the two tables
   is_changed     ;whether the user has changed the record

;  Changes made to the ENROLL2 record must also be made to the
;  companion ENROLL1 record.

;  If the user presses the [Del] key, there is no special processing.
;  We simply delete the record.

If user_key="Del" then     ;the [Del] key
   Return True             ;no special processing or testing to do
Endif

;  Perform standard record departure on the record

is_changed = recordstatus("Modified")  ;has the user made changes?
Rec_Dep_Std()              ;standard record departure procedure
If not retval then         ;it failed
   Return False            ;do not proceed
Endif
If not is_posted_key or not is_changed then  ;is there some new change?
   Return True                               ;no, nothing more to do
Endif

;  Update the master table, and thus any other links,
;  with the new source table key

CopyToArray enr_rec              ;capture the new or changed record
current_student = [Student ID]   ;find out the current student
student_rec[1] = ""              ;the record array is out of date
MoveTo "Master"                  ;we need to update the master record...
[Student ID] = current_student   ;...with the student # key

;  Transfer the record to the ENROLL1 table

MoveTo "Enroll1"                             ;move to the companion table
enr_rec["Student ID"] = enr_rec["Class ID"]  ;swap the two key fields
enr_rec["Class ID"]   = current_student
Ins                                          ;insert a new record
CopyFromArray enr_rec                        ;fill in the record
LockRecord                                   ;post it
If not retval then                           ;the lock failed
   If errorcode()<>53 then                   ;not a duplicate key
      wait_msg = "Error updating the ENROLL1 table:  "+errormessage()
      MoveTo "Enroll2"
      Return False
   Endif
Endif
UnLockRecord                                 ;post the changes

MoveTo "Enroll2"     ;return to the original table

Return True          ;successful completion

EndProc

;  ------------------------------------------------------------------

Proc Move_Shad_Stu()

;  Move to the next record in the shadow student table table.

;  Special record movement key:  We simulate record movement
;  in a "shadow" table, based on the key pressed by the user.

MoveTo "STUDENT"

If not is_editing then           ;we're in a display-only mode
   ImageRights ReadOnly          ;that will affect how some keys work
Endif

Switch   ;convert form-view movement to table-view movement
   Case user_key="PgUp":      Up
   Case user_key="PgDn":      Down
   Otherwise:                 KeyPress user_key
EndSwitch

ImageRights                      ;restore full image rights
current_student = [Student ID]   ;find out the current student
CopyToArray student_rec          ;fetch the entire student record

Page_Arrival_2()                 ;update the records on page 2

EndProc

;  ------------------------------------------------------------------

Proc Move_Shad_Clas()

;  Move to the next record in the shadow class table table.

;  Special record movement key:  We simulate record movement
;  in a "shadow" table, based on the key pressed by the user.

MoveTo "CLASS"

If not is_editing then           ;we're in a display-only mode
   ImageRights ReadOnly          ;that will affect how some keys work
Endif

Switch   ;convert form-view movement to table-view movement
   Case user_key="PgUp":      Up
   Case user_key="PgDn":      Down
   Otherwise:                 KeyPress user_key
EndSwitch

ImageRights                   ;restore full image rights
current_class = [Class ID]    ;find out the current class
CopyToArray class_rec         ;fetch the entire class record

Page_Arrival_4()              ;update the records on page 4

EndProc

;  ------------------------------------------------------------------

Proc Rec_Del_Enroll1()

;  Process the "Del" key for the enrollment #1 table.
;  On a "Del" key, locate and delete the companion record

;  Update the master table, and thus any other links,
;  with the new source table key

current_student = [Student ID]   ;find out the current student
student_rec[1] = ""              ;the record array is out of date
current_class = [Class ID]       ;find out the current class
class_rec[1]  = ""               ;no class record in an array yet
MoveTo "Master"                  ;we need to update the master record...
[Class ID] = current_class       ;...with the class # key

;  Locate and delete the record in the ENROLL2 table

MoveTo "Enroll2"              ;move to the companion table
MoveTo [Student ID]           ;the field to search on
Locate current_student        ;find the companion record
If retval then                ;we found it
   Del                        ;so delete it
Endif

;  Delete the original record

MoveTo "Enroll1"     ;return to the original table
Del

EndProc

;  ------------------------------------------------------------------

Proc Rec_Del_Enroll2()

;  Process the "Del" key for the enrollment #2 table.
;  On a "Del" key, locate and delete the companion record

;  Update the master table, and thus any other links,
;  with the new source table key

current_class = [Class ID]       ;find out the current class
class_rec[1]  = ""               ;no class record in an array yet
current_student = [Student ID]   ;find out the current student
student_rec[1] = ""              ;the record array is out of date
MoveTo "Master"                  ;we need to update the master record...
[Student ID] = current_student   ;...with the student # key

;  Locate and delete the record in the ENROLL1 table

MoveTo "Enroll1"              ;move to the companion table
MoveTo [Class ID]             ;the field to search on
Locate current_class          ;find the companion record
If retval then                ;we found it
   Del                        ;so delete it
Endif

;  Delete the original record

MoveTo "Enroll2"     ;return to the original table
Del

EndProc

;  ------------------------------------------------------------------

Proc Del_Confirm()

;  Displays a menu when the user presses the [Del] key during
;  a WAIT command.

;  Returns retval=True if the record is to be deleted 
;  and retval=False if it is not.

Private wait_menu_sel

Beep

If not is_editing then
   Return False
Endif

Message "Do you really want to delete this record?"
ShowMenu
   "No":       "Do not delete this record",
   "Yes":      "Go ahead and delete this record"
   to wait_menu_sel

Message ""
@ 1,0 Clear Eol
@ 0,0 Clear Eol

If wait_menu_sel<>"Yes" then Return False Endif

Return True

EndProc

;  ------------------------------------------------------------------

Proc Zoom_Echo(echo_off_after)

;  Simulates Ctrl-Z using ECHO

;  Parameter echo_off_after -- True or False; whether to turn
;                              "Echo" off at the conclusion of
;                              this procedure

Private 
   k,       ;the keystroke which has been pressed
   w        ;the window error message, if any

;  Press the Ctrl-Z key

Zoom                 ;press [Ctrl-Z] for the user

;  Read each key from the keyboard and press it, one at a time.
;  Set ECHO NORMAL so the user can see their interaction with the
;  Paradox workspace.
;  Keep processing keystrokes until the cursor departs from the
;  "Value:" prompt at the top of the screen.

While True           ;loop for each keystroke the user presses

;  Use the MENUCHOICE() function to determine if the cursor is
;  on the top line of the screen.  If it isn't, then we're done
;  zooming.

   If menuchoice()="Error" then  ;no more Ctrl-Z prompt, probably
      w = window()         ;capture error message, if any
      Echo Off             ;lower canvas to do SyncCursor
      SyncCursor           ;move canvas cursor to workspace
      Echo Normal          ;turn echo back off
      If row()<>0 then     ;make sure we're no longer on top row
         QuitLoop          ;which means we're all done zooming
      Endif
   Endif

;  Read a character from the keyboard

   Echo Normal             ;show the user everything that happens
   k = getchar()                 ;let the user press a key

;  Certain keys require special processing

   If (-68<=k and k<=-59)        ;trap [F1] thru [F10]
      or (-113<=k and k<=-84 and k<>-108) ;trap other fcn keys
      or k=15 or k=-24 then      ;trap [Ctrl-O] and [Alt-O]
         Beep                    ;beep on a bad keypress

;  Most keys simply get passed through as-is to Paradox

   Else                          ;it's an okay key
      KeyPress k                 ;so let the user press it
   Endif
EndWhile                   ;do the next key

;  All done.  Quit echoing.

If echo_off_after then     ;we're supposed to stop echoing
   Echo Off Endif          ;so turn off echo
Return w                   ;all done; return final message

EndProc

;  ------------------------------------------------------------------

Main()         ;perform the main procedure

;  ------------------------------------------------------------------

